/************************************************
* RELFILE - relative file manager toolbox.		*
*												*
* 900422 MCMason - Microsoft C V5.10			*
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include "files.h"

#define TRUE  (1==1)
#define FALSE (1==0)

static int  _ErrNum=fOK;
static char _RecStat;

#define rf_HDR_SIZE 20	/* size of file header	*/

/************************************************
* _rfSeek - check if file type is valid, and	*
* set file pointer to correct position for rec	*
*												*
* INP:	f - the file to seek the record in		*
*		n - the record number to seek			*
* OUT:	Zero (FALSE) if no error, otherwise the	*
*		error code for the error encountered.	*
************************************************/
int _rfSeek(RELFILE f,long n)
	{
	long pos;

	/* Did we get a relative file? */
	if (f->FileType != fRELFILE)
		{
		_ErrNum = fBADFILETYPE;
		return _ErrNum;
		}

	/* Set file position to record position */
	if (n > f->LastRec)
		{
		_ErrNum = fBADRECNO;
		return _ErrNum;
		}
	pos = (n-1)*(f->RecSize+1) + rf_HDR_SIZE;
	if (lseek(f->File,pos,SEEK_SET)!=pos)
		{
		_ErrNum = fSEEK;
		return _ErrNum;
		}
	if (read(f->File,&_RecStat,sizeof(char)) !=
			sizeof(char))
		{
		_ErrNum = fREAD;
		return _ErrNum;
		}
	return fOK;
	}

/************************************************
* _rfMark - mark the record type (deleted, or	*
* active)										*
*												*
* INP:	f - the file holding the record			*
*		n - the record to be marked				*
* OUT:	TRUE on error (use rfLastError to get	*
*		the error code), FALSE otherwise.		*
* NOTE:	It is assumed that any required error	*
*		checking has already been performed, so	*
*		the absolute minimum is in this routine	*
************************************************/
int _rfMark(RELFILE f,long n,int mark)
	{
	long pos;
	char mk;

	/* Set file position to record position */
	if (n > f->LastRec)
		{
		_ErrNum = fBADRECNO;
		return _ErrNum;
		}
	pos = (n-1)*(f->RecSize+1) + rf_HDR_SIZE;
	if (lseek(f->File,pos,SEEK_SET)!=pos)
		{
		_ErrNum = fSEEK;
		return _ErrNum;
		}
	mk=(char)mark;
	if (write(f->File,&mk,sizeof(char)) !=
			sizeof(char))
		{
		_ErrNum = fWRITE;
		return _ErrNum;
		}
	return fOK;
	}

/************************************************
* rfCreate - create a relative file				*
*												*
* INP:	n - file name (full path name)			*
*		s - record size (4..65535 bytes)		*
* OUT:	0 (fOK) if no errors found, otherwise,	*
*		the appropriate error code is returned	*
************************************************/
int rfCreate(char *n, unsigned s)
	{
	struct stat sb;
	struct RelFileDescr rfd;
	
	/* Check if file exists, or is dir name */
	if (!stat(n,&sb))
		if (sb.st_mode & S_IFDIR)
			return fDIR;
		else
			return fFILE;

	/* Create the file */
	rfd.File = open(n,O_CREAT|O_WRONLY|O_BINARY,
				S_IREAD|S_IWRITE);
	if (rfd.File<0) return fCREATE;

	/* Create the file header */
	rfd.FileType = fRELFILE;
	rfd.RecSize  = s;
	rfd.RecsUsed = 0L;
	rfd.RecsFree = 0L;
	rfd.NextRec  = 0L;
	rfd.LastRec  = 0L;

	/* write the header */
	if (write(rfd.File,(char *)&rfd,rf_HDR_SIZE)
				!= rf_HDR_SIZE)
		return fWRITE;
	else if (close(rfd.File))
		return fCLOSE;
	else
		return fOK;
	}

/************************************************
* rfDestroy - delete a relative file			*
*												*
* INP:	n - name of the file to delete			*
* OUT:	TRUE if error, FALSE otherwise, You can	*
*		call rfLastError to retrieve the code	*
************************************************/
int rfDestroy(char far *n)
	{
	RELFILE s;
	
	/* Let rfOpen & rfClose do the dirty work	*
	 * of checking file validity				*/
	s=rfOpen(n);
	if (s==NULL)
		return TRUE;
	if (rfClose(s)!=fOK)
		return TRUE;

	/* The file passes inspection ... delete it	*/
	unlink(n);
	return fOK;
	}

/************************************************
* rfOpen - open the relative file				*
*												*
* INP:	n - the name of the file to open		*
* OUT:	NULL - the file could not be opened, or	*
*		a pointer to RELFILE.					*
************************************************/
RELFILE rfOpen(char *n)
	{
	struct stat sb;
	RELFILE r;
	
	/* Check if file exists */
	if (stat(n,&sb))
		{
		_ErrNum = fDNE;
		return (RELFILE)NULL;
		}

	/* Is it a subdirectory? */
	if (sb.st_mode & S_IFDIR)
		{
		_ErrNum = fDIR;
		return (RELFILE)NULL;
		}

	/* Get a node to stick the block in */
	r = malloc(sizeof(struct RelFileDescr));
	if (r==NULL)
		{
		_ErrNum = fNORAM;
		return NULL;
		}
	
	/* Is it a relative file? */
	r->File=open(n,O_RDWR|O_BINARY);
	if (r->File==-1)
		{
		_ErrNum = fOPEN;
		return (RELFILE)NULL;
		}
	if (read(r->File,(char *)r,rf_HDR_SIZE)
				!=rf_HDR_SIZE)
		{
		_ErrNum = fREAD;
		return (RELFILE)NULL;
		}
	if (r->FileType != fRELFILE)
		{
		_ErrNum = fBADFILETYPE;
		return (RELFILE)NULL;
		}
	return r;
	}

/************************************************
* rfClose - close the relative file				*
*												*
* INP:	f - the RELFILE to close				*
* OUT:	TRUE on error, FALSE otherwise			*
************************************************/
int rfClose(RELFILE f)
	{
	/* Check if the file type is valid */
	if (f->FileType != fRELFILE)
		{
		_ErrNum = fBADFILETYPE;
		return TRUE;
		}
	
	/* Write out the modified control block */
	if (lseek(f->File,0L,SEEK_SET)!=0L)
		{
		_ErrNum = fSEEK;
		return TRUE;
		}
	if (write(f->File,(char *)f,rf_HDR_SIZE)
				!=rf_HDR_SIZE)
		{
		_ErrNum = fWRITE;
		return TRUE;
		}

	/* Close the file */
	if (close(f->File))
		{
		_ErrNum = fCLOSE;
		return TRUE;
		}
		
	/* Invalidate the block's file type */
	f->FileType = fBADFILETYPE;
	return FALSE;
	}
	
/************************************************
* rfGetRec - get a new record (add a new one to	*
* the end of the file if necessary).			*
*												*
* INP:	f - the RELFILE to get a block from		*
* OUT:	the number (key) of the new block, or	*
*		0 on error.								*
************************************************/
long rfGetRec(RELFILE f)
	{
	long retval;
	
	/* Check if the file type is valid */
	if (f->FileType != fRELFILE)
		{
		_ErrNum = fBADFILETYPE;
		return 0L;
		}

	/* Any records on free list? */
	if (f->NextRec != 0L)
		{	/* Yes, remove from free list & return it */
		retval = f->NextRec;
		if (_rfSeek(f,retval))
			return 0L;
		if (_RecStat!=fDELETED)
			{
			_ErrNum = fBADREQ;
			return 0L;
			}
		if (read(f->File,(char *)&(f->NextRec),
				sizeof(long))!=sizeof(long))
			{
			_ErrNum = fBADFILE;
			return 0L;
			}
		f->RecsFree--;
		}
	else
		{	/* No, add one to end of file	*/
		f->LastRec++;
		if (chsize(f->File,rf_HDR_SIZE+
				f->LastRec*(f->RecSize+1)))
			{
			_ErrNum = fEXTEND;
			return 0L;
			}
		retval = f->LastRec;
		}

	/* Mark the record as ACTIVE */
	if (_rfMark(f,retval,fACTIVE))
		return 0L;
	return retval;
	}

/************************************************
* rfDelRec - delete a record (if the record is	*
* at the end of the file, truncate the file)	*
*												*
* INP:	f - the RELFILE of the record to dispose*
*		n - the number of the record to dispose	*
* OUT:	TRUE on error, FALSE otherwise			*
************************************************/
int rfDelRec(RELFILE f, long n)
	{
	if (_rfSeek(f,n)) return TRUE;
	if (_RecStat != fACTIVE)
		{
		_ErrNum = fBADREQ;
		return TRUE;
		}
	if (n==f->LastRec)
		{
		--f->LastRec;
		if (chsize(f->File,rf_HDR_SIZE+
				f->LastRec*(f->RecSize+1)))
			{
			_ErrNum = fEXTEND;
			return TRUE;
			}
		}
	else
		{
		if (_rfMark(f,n,fDELETED))
			return TRUE;
		if (write(f->File,(char *)&(f->NextRec),
				sizeof(long))!=sizeof(long))
			{
			_ErrNum = fBADFILE;
			return TRUE;
			}
		f->NextRec = n;
		++f->RecsFree;
		}
	--f->RecsUsed;
	return FALSE;
	}

/************************************************
* rfRead - read the contents of a file block	*
*												*
* INP:	f - the RELFILE holding the record		*
*		n - the record number (key)				*
*		buf - the address of the buffer that	*
*			will hold the record just read.		*
* OUT:	TRUE on error, FALSE otherwise			*
************************************************/
int rfRead(RELFILE f, long n, void *buf)
	{
	/* Verify file & seek to proper position */
	if (_rfSeek(f,n) != fOK) return TRUE;
	if (_RecStat != fACTIVE)
		{
		_ErrNum = fBADREQ;
		return TRUE;
		}

	/* Read the record from the file */
	if (read(f->File,buf,f->RecSize)!=f->RecSize)
		{
		_ErrNum = fREAD;
		return TRUE;
		}
	return FALSE;
	}

/************************************************
* rfWrite - write a record to the file			*
*												*
* INP:	f - the RELFILE of the block to write	*
*		n - the record number to write			*
* 		buff - the address of the buffer that	*
*			holds the data to write				*
* OUT:	TRUE on error, FALSE otherwise			*
************************************************/
int rfWrite(RELFILE f, long n, void *buf)
	{
	/* Verify file & seek to proper position */
	if (_rfSeek(f,n) != fOK) return TRUE;
	if (_RecStat != fACTIVE)
		{
		_ErrNum = fBADREQ;
		return TRUE;
		}

	/* Write the record to the file */
	if (write(f->File,buf,f->RecSize)!=f->RecSize)
		{
		_ErrNum = fWRITE;
		return TRUE;
		}
	return FALSE;
	}

/************************************************
* rfRecStat - find the status of a record		*
*												*
* INP:	f - the file holding the record			*
*		n - the record of interest				*
* OUT:	If the record is deleted, return the	*
*		number of the next record in the free	*
*		list (0 means end of free list).  Active*
*		records get a -1, errors get a -2		*
************************************************/
long rfRecStat(RELFILE f, long n)
	{
	long retval;
	/* Verify file & seek to proper position */
	if (_rfSeek(f,n) != fOK) return -2L;
	if (_RecStat == fACTIVE) return -1L;
	/* Get the next record # in free list	*/
	if (read(f->File,(char *)&retval,sizeof(long))
			!= sizeof(long))
		{
		_ErrNum = fWRITE;
		return -2L;
		}
	return retval;
	}

/************************************************
* rfLastError - return last error encountered	*
*												*
* OUT:	fOK if no errors have occurred since	*
*		the last call, otherwise the last error	*
*		found.									*
************************************************/
int rfLastError(void)
	{
	int i;
	i=_ErrNum;
	_ErrNum=fOK;
	return i;
	}
